package at.ac.uibk.gitsearch.repository.search;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.MultiMatchQueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.search.fetch.subphase.highlight.HighlightBuilder;
import org.springframework.stereotype.Repository;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;

import at.ac.uibk.gitsearch.config.ApplicationProperties;
import at.ac.uibk.gitsearch.service.dto.SearchInputDTO;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO.GitProject;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO.Person;
import at.ac.uibk.gitsearch.service.dto.SearchResultsDTO.SearchResultDTO;

@Repository
public class MetaDataRepository {

	/**
	 * just a local wrapper, to decrypt JSON
	 *
	 */
	public static class SearchResultDTOWrapper {
		private SearchResultDTO metadata = new SearchResultDTO();

		public SearchResultDTO getMetadata() {
			return metadata;
		}

		public void setMetadata(SearchResultDTO metadata) {
			this.metadata = metadata;
		}

		private GitProject project;

		public GitProject getProject() {
			return project;
		}

		public void setProject(GitProject project) {
			this.project = project;
		}



	}

	private final RestHighLevelClient elasticsearchClient;

	private final ApplicationProperties properties;

	private final int FRAGMENT_SIZE = 5000;

	public MetaDataRepository(RestHighLevelClient elasticsearchClient, ApplicationProperties properties) {
		this.elasticsearchClient = elasticsearchClient;
		this.properties = properties;

	}

	/**
	 * contains for each Field, a list of Strings mapped to a completion.
	 * E.g. "metadata.creator" -> Kerstin -> {"Kerstin Limbeck", "Kerstin Mair"}
	 */
	private Map<String, Map<String, Set<String>>> cachedCompletions = null;

	private static final int MAX_AUTO_COMPLETION_RESULTS = 10;

	/**
	 * returns all keyword autocompletes for keyWord
	 *
	 * @param keyWordPrefix
	 * @return
	 * @throws IOException
	 */
	public List<String> getKeywordsAutoComplete(String keyWordPrefix) throws IOException {
		return getFieldAutoCompletion(keyWordPrefix, SearchRepositoryConstants.METADATA_KEYWORDS);
	}

	/**
	 * returns all programmingLanguage autocompletes for keyWord
	 *
	 * @param progammingLanguagePrefix
	 * @return
	 * @throws IOException
	 */
	public List<String> getProgrammingLanguageAutoComplete(String progammingLanguagePrefix) throws IOException {
		return getFieldAutoCompletion(progammingLanguagePrefix, SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES);
	}

	/**
	 * returns all creator autocompletes
	 *
	 * @param creatorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<String> getCreatorAutoComplete(String creatorPrefix) throws IOException {
		return getFieldAutoCompletion(creatorPrefix, SearchRepositoryConstants.METADATA_CREATOR);
	}

	/**
	 * returns all contributor autocompletes
	 *
	 * @param contributorPrefix
	 * @return
	 * @throws IOException
	 */
	public List<String> getContributorAutoComplete(String contributorPrefix) throws IOException {
		return getFieldAutoCompletion(contributorPrefix, SearchRepositoryConstants.METADATA_CONTRIBUTOR);
	}

	private List<String> getFieldAutoCompletion(String keyWordPrefix, final String metaDataField)
			throws IOException, JsonProcessingException, JsonMappingException {
		fillAutoCompletion();
		return cachedCompletions.get(metaDataField).entrySet().stream()
				.filter(s -> s.getKey().startsWith(keyWordPrefix))
				.map(Map.Entry::getValue)
				.flatMap(java.util.Collection::stream).distinct()
				.limit(MAX_AUTO_COMPLETION_RESULTS).collect(Collectors.toList());
	}

	private synchronized void fillAutoCompletion() throws IOException, JsonProcessingException, JsonMappingException {
		if (cachedCompletions == null) {
			cachedCompletions = new HashMap<>();
			cachedCompletions.put(SearchRepositoryConstants.METADATA_KEYWORDS, new TreeMap<>());
			cachedCompletions.put(SearchRepositoryConstants.METADATA_CREATOR, new TreeMap<>());
			cachedCompletions.put(SearchRepositoryConstants.METADATA_CONTRIBUTOR, new TreeMap<>());
			cachedCompletions.put(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES, new TreeMap<>());

			SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);

			SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);
			ObjectMapper objectMapper = new ObjectMapper();
			objectMapper.registerModule(new JavaTimeModule());
			objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

			for (SearchHit searchHit : searchResponse.getHits().getHits()) {
				final String sourceAsJSON = searchHit.getSourceAsString();
				SearchResultDTOWrapper entry = objectMapper.readValue(sourceAsJSON, SearchResultDTOWrapper.class);
				if (entry.getMetadata().getKeyword() != null)
					for (String keyWord : entry.getMetadata().getKeyword())
						addTo(cachedCompletions.get(SearchRepositoryConstants.METADATA_KEYWORDS), split(keyWord), keyWord);
				if (entry.getMetadata().getContributor() != null)
					for (Person contributor : entry.getMetadata().getContributor())
						addTo(cachedCompletions.get(SearchRepositoryConstants.METADATA_CONTRIBUTOR), split(contributor.getName()), contributor.getName());
				if (entry.getMetadata().getCreator() != null)
					for (Person creator : entry.getMetadata().getCreator())
						addTo(cachedCompletions.get(SearchRepositoryConstants.METADATA_CREATOR),
								split(creator.getName()), creator.getName());
				if (entry.getMetadata().getProgrammingLanguage() != null)
					for (String language : entry.getMetadata().getProgrammingLanguage())
						addTo(cachedCompletions.get(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES), Collections.singletonList(language), language);
			}
		}
	}

	private void addTo(Map<String, Set<String>> completionMap, List<String> tokenList, String target) {
		tokenList.forEach(token -> {
			Set<String> existingTargets = completionMap.get(token);
			if(existingTargets == null) {
				existingTargets = new HashSet<>();
				completionMap.put(token,  existingTargets);
			}
			existingTargets.add(target);
		});
	}



	private List<String> split(String s) {
		StringTokenizer st = new StringTokenizer(s, " \t\n\r\f ");
		List<String> result = new ArrayList<>();
		while (st.hasMoreTokens()) {
			result.add(st.nextToken());
		}
		return result;
	}

	public SearchResultsDTO pageDetails(SearchInputDTO searchInputDTO, int pageSize) throws IOException {
		fillAutoCompletion();
		SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);

		BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();
		if(!StringUtils.isEmpty(searchInputDTO.getFulltextQuery()))
			queryBuilder.must(QueryBuilders.multiMatchQuery(
						searchInputDTO.getFulltextQuery(),
						SearchRepositoryConstants.METADATA_DESCRIPTION,
						SearchRepositoryConstants.METADATA_KEYWORDS,
						SearchRepositoryConstants.METADATA_TITLE)
						.type(MultiMatchQueryBuilder.Type.PHRASE_PREFIX)
						);

		if(!StringUtils.isEmpty(searchInputDTO.getMetadata().getProgrammingLanguage())) {
			queryBuilder.must(QueryBuilders.simpleQueryStringQuery(searchInputDTO.getMetadata().getProgrammingLanguage()).field(SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES));
		}

		if(!StringUtils.isEmpty(searchInputDTO.getMetadata().getKeyword())) {
			queryBuilder.must(QueryBuilders.simpleQueryStringQuery(searchInputDTO.getMetadata().getKeyword()).field(SearchRepositoryConstants.METADATA_KEYWORDS));
		}

		if(!StringUtils.isEmpty(searchInputDTO.getMetadata().getAuthor())) {
			queryBuilder.must(QueryBuilders.simpleQueryStringQuery(searchInputDTO.getMetadata().getAuthor()).field(SearchRepositoryConstants.METADATA_CREATOR));
		}

		if(!StringUtils.isEmpty(searchInputDTO.getMetadata().getLicense())) {
			queryBuilder.must(QueryBuilders.simpleQueryStringQuery(searchInputDTO.getMetadata().getLicense()).field(SearchRepositoryConstants.METADATA_LICENSE));
		}

        char[] boundaryChars = {'\n'};

        HighlightBuilder highlightBuilder = new HighlightBuilder()
            .highlighterType("plain")
            .preTags(properties.getSearch().getHighlightPre())
            .postTags(properties.getSearch().getHighlightPost())
            .field(SearchRepositoryConstants.METADATA_DESCRIPTION)
            .boundaryChars(boundaryChars)
            .boundaryScannerType(HighlightBuilder.BoundaryScannerType.SENTENCE)
            .fragmentSize(FRAGMENT_SIZE);

        // String[] includes = {};
		// String[] excludes = {HighlightRepositoryConstants.CONTENT_FIELD};

		SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
		sourceBuilder.query(queryBuilder)
            .highlighter(highlightBuilder)
				.size(pageSize).from((searchInputDTO.getPage()) * pageSize);
		// .fetchSource(includes, excludes);

		searchRequest.source(sourceBuilder);

		SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

		ObjectMapper objectMapper = new ObjectMapper();
		objectMapper.registerModule(new JavaTimeModule());
		objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

		final SearchResultsDTO results = new SearchResultsDTO();
		List<SearchResultDTO> searchResults = new ArrayList<>();
		for (SearchHit searchHit : searchResponse.getHits().getHits()) {
			SearchResultDTOWrapper entry = objectMapper.readValue(searchHit.getSourceAsString(), SearchResultDTOWrapper.class);
			final SearchResultDTO metadata = entry.getMetadata();
			metadata.setProject(entry.getProject());
			searchResults.add(metadata);
		}
		long hitCount = searchResponse.getHits().getTotalHits();
		results.setHitCount(hitCount);
		results.setSearchResult(searchResults);
		return results;
	}

}
